/* Set up a dialog to convert a geographic location in
   between decimal degrees, degrees with decimal minutes,
   and degrees, minutes and decimal seconds.

   After each conversion, a hyperlink is presented to
   maps.google.com, where this location can be viewed
   on satellite pictures and maps.

   15 April 2023:
   Added support for the Maidenhead Locator System.
 */

#include <stdio.h>
#include <stdlib.h>
#if !defined(_Windows) && !defined (__MSDOS__)
  #include <sys/types.h>
  #include <unistd.h>
#endif
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>

#include "../cgitools/cgitools.h"

#if !defined Minimum
   #define Minimum(a,b) (a) > (b) ? (b) : (a)
#endif

struct progdata_s
{
   FILE *errfp;
   char lang[2+1];
   char page[4+1];
   char *scriptname;
} progdata;

enum algor_enum
{
   ALGOR_deg, /* decimal degrees as input, calculate the other ones */
   ALGOR_dmm, /* degrees and decimal minutes as input, calculate the other ones */
   ALGOR_dms, /* degrees, minutes and decimal seconds as input, calculate the other ones */
   ALGOR_mdh, /* Maidenhead Locator System */
};
struct formdata_s
{
   char lat_deg_degrees[16];

   char lat_dmm_degrees[10];
   char lat_dmm_minutes[15];

   char lat_dms_degrees[10];
   char lat_dms_minutes[10];
   char lat_dms_seconds[15];

   char lon_deg_degrees[16];

   char lon_dmm_degrees[10];
   char lon_dmm_minutes[15];

   char lon_dms_degrees[10];
   char lon_dms_minutes[10];
   char lon_dms_seconds[15];

   char NorthSouth[2];
   char EastWest  [2];

   char Maidenhead[8 + 4 + 1];

   char GoogleMapsUrl[75];
   enum algor_enum Algor;
} formdata;

static const double Maidenhead_factors[] =
{
   20.0,
   10.0,
   20.0 / 10.0,
   10.0 / 10.0,
   20.0 / 10.0 / 24.0,
   10.0 / 10.0 / 24.0,
   20.0 / 10.0 / 24.0 / 10.0,
   10.0 / 10.0 / 24.0 / 10.0,
   20.0 / 10.0 / 24.0 / 10.0 / 24.0,
   10.0 / 10.0 / 24.0 / 10.0 / 24.0,
   20.0 / 10.0 / 24.0 / 10.0 / 24.0 / 10.0,
   10.0 / 10.0 / 24.0 / 10.0 / 24.0 / 10.0,
};

void get_invocation_parms (
   struct progdata_s *p_progdata,
   struct formdata_s *p_formdata);

void showform (
   struct progdata_s *p_progdata,
   struct formdata_s *p_formdata);

void getparms (int len, struct formdata_s *p_formdata);

int initialise (struct formdata_s *p_formdata);

int calculate (struct formdata_s *p_formdata);

int FlipMinus (double degrees, int *p_dir);

void ValidateMaidenhead  (struct formdata_s *p_formdata);
void CalculateMaidenhead (struct formdata_s *p_formdata);

static char ErrorMessageBuffer[512];
static void ErrorOnScreen (void)
{
   printf("Content-type: text/html\n\n");
   printf("<html>\n");
   printf("<p>%s</p>\n", ErrorMessageBuffer);
   printf("</html>\n");

   fflush(stdout);
}

int main (int argc, char **argv)
{
   int len;
   char *ErrFile = "degmapgu/degmapgu.err";
   char *ErrFileMode = "at";
   progdata.scriptname = "../cgi-bin/degmapgu.cgi";

   progdata.errfp = fopen(ErrFile, ErrFileMode);
   if (!progdata.errfp)
   {
       strncpy(ErrorMessageBuffer, strerror(errno), sizeof ErrorMessageBuffer - 1);
       #if !defined(_Windows) && !defined (__MSDOS__)
       sprintf(ErrorMessageBuffer + strlen(ErrorMessageBuffer),
          "<br>Cannot open file %s in mode \"%s\", <br>user/grp = %d/%d, eff. user/grp = %d/%d\n",
          ErrFile, ErrFileMode, getuid(), geteuid(), getgid(), getegid());
       #else
       sprintf(ErrorMessageBuffer + strlen(ErrorMessageBuffer),
          "<br>Cannot open file %s in mode \"%s\"\n",
          ErrFile, ErrFileMode);
       #endif
       ErrorOnScreen();
       return 1;
   }
   get_invocation_parms(&progdata, &formdata);

   if ((len = userinput()) == 0)
   {
      /* Initialise values */
      initialise(&formdata);

      /* Calculate other values */
      if (formdata.Maidenhead[0])
         formdata.Algor = ALGOR_mdh;
      else
         formdata.Algor = ALGOR_dms;
      calculate(&formdata);
      showform(&progdata, &formdata);
   }
   else
   {
      getparms(len, &formdata);
      calculate(&formdata);
      showform(&progdata, &formdata);
   }

   fclose(progdata.errfp);

   return 0;
}

void get_invocation_parms (
   struct progdata_s *p_progdata,
   struct formdata_s *p_formdata)
{
   char *query;
   char **Vars = NULL, **V;

   memset(p_formdata, '\0', sizeof *p_formdata);

   strcpy(p_progdata->lang, "en");
   strcpy(p_progdata->page, "text");

   query = getenv("QUERY_STRING");
   if (query)
   {
      Vars = CgiConvVars(query, NULL);

      for (V = Vars; V[0]; V += 2)
      {
         if (strcmp(V[0], "lang") == 0 && strlen(V[1]))
         {
            sprintf(p_progdata->lang, "%2.2s", V[1]);
         }
         else if (strcmp(V[0], "page") == 0 && strlen(V[1]))
         {
            sprintf(p_progdata->page, "%4.4s", V[1]);
         }
         else if (strcmp(V[0], "loc") == 0 && strlen(V[1]))
         {
            sprintf(p_formdata->Maidenhead, "%-12.12s", V[1]);
         }
      }
   }
}

#define CHUNK 10

void showform (
   struct progdata_s *p_progdata,
   struct formdata_s *p_formdata)
{
   char htmlfile[41];
   FILE *fp;
   int bufsize = CHUNK;
   char *buf;
   char *b;
   char c;

   sprintf(htmlfile, "degmapgu/%.2s/dmg-%.4s.htm",
      p_progdata->lang, p_progdata->page);

   fp = fopen(htmlfile, "rt");
   if (!fp)
   {
      time_t now;
      fprintf(p_progdata->errfp, "%s", ctime(((now = time(NULL)), &now)));
      fprintf(p_progdata->errfp,
         "Cannot open file %s, errno=%d = %s\n",
         htmlfile, errno, strerror(errno));
      return;
   }

   buf = malloc(bufsize);
   if (!buf)
   {
      time_t now;
      fprintf(p_progdata->errfp, "%s", ctime(((now = time(NULL)), &now)));
      fprintf(p_progdata->errfp, "Cannot malloc %d bytes\n", bufsize);
      return;
   }
   b = buf;

   while ((c = getc(fp)) != EOF)
   {
      if (b - buf >= bufsize)
      {
         buf = realloc(buf, bufsize + CHUNK);
         if (!buf)
         {
            time_t now;
            fprintf(p_progdata->errfp, "%s", ctime(((now = time(NULL)), &now)));
            fprintf(p_progdata->errfp,
               "Cannot malloc %d bytes\n", bufsize + CHUNK);
            return;
         }
         b = buf + bufsize;
         bufsize += CHUNK;
      }
      *b++ = (char)c;
   }
   *b = '\0';

   fclose(fp);

   printf("Content-type: text/html\n\n");
   printf(buf, p_progdata->scriptname, p_progdata->lang,
      p_formdata->NorthSouth[0] == 'N' ? "Selected" : "",
      p_formdata->NorthSouth[0] == 'S' ? "Selected" : "",
      p_formdata->EastWest  [0] == 'E' ? "Selected" : "",
      p_formdata->EastWest  [0] == 'W' ? "Selected" : "",

      p_formdata->lat_deg_degrees,

      p_formdata->lon_deg_degrees,

      p_formdata->lat_dmm_degrees,
      p_formdata->lat_dmm_minutes,

      p_formdata->lon_dmm_degrees,
      p_formdata->lon_dmm_minutes,

      p_formdata->lat_dms_degrees,
      p_formdata->lat_dms_minutes,
      p_formdata->lat_dms_seconds,

      p_formdata->lon_dms_degrees,
      p_formdata->lon_dms_minutes,
      p_formdata->lon_dms_seconds,

      p_formdata->Maidenhead,

      p_formdata->GoogleMapsUrl);

   fflush(stdout);
}

void getparms (int len, struct formdata_s *p_formdata)
{
   char **Vars = NULL, **V;
   char *buf = calloc(len + 1, 1);

   if (!buf ||
      fread(buf, len, 1, stdin) != 1)
   {
      free(buf);
      return;
   }

   Vars = CgiConvVars(buf, NULL);

   for (V = Vars; V[0]; V += 2)
   {
      if (strcmp(V[0], "lat_deg_degrees") == 0 && strlen(V[1]))
      {
         sprintf(p_formdata->lat_deg_degrees, "%-15.15s", V[1]);
      }
      else if (strcmp(V[0], "lon_deg_degrees") == 0 && strlen(V[1]))
      {
         sprintf(p_formdata->lon_deg_degrees, "%-15.15s", V[1]);
      }

      else if (strcmp(V[0], "lat_dmm_degrees") == 0)
      {
         sprintf(p_formdata->lat_dmm_degrees, "%-4.4s", V[1]);
      }
      else if (strcmp(V[0], "lon_dmm_degrees") == 0)
      {
         sprintf(p_formdata->lon_dmm_degrees, "%-4.4s", V[1]);
      }
      else if (strcmp(V[0], "lat_dmm_minutes") == 0)
      {
         sprintf(p_formdata->lat_dmm_minutes, "%-10.10s", V[1]);
      }
      else if (strcmp(V[0], "lon_dmm_minutes") == 0)
      {
         sprintf(p_formdata->lon_dmm_minutes, "%-10.10s", V[1]);
      }

      else if (strcmp(V[0], "lat_dms_degrees") == 0)
      {
         sprintf(p_formdata->lat_dms_degrees, "%-4.4s", V[1]);
      }
      else if (strcmp(V[0], "lon_dms_degrees") == 0)
      {
         sprintf(p_formdata->lon_dms_degrees, "%-4.4s", V[1]);
      }
      else if (strcmp(V[0], "lat_dms_minutes") == 0)
      {
         sprintf(p_formdata->lat_dms_minutes, "%-4.4s", V[1]);
      }
      else if (strcmp(V[0], "lon_dms_minutes") == 0)
      {
         sprintf(p_formdata->lon_dms_minutes, "%-4.4s", V[1]);
      }
      else if (strcmp(V[0], "lat_dms_seconds") == 0)
      {
         sprintf(p_formdata->lat_dms_seconds, "%-10.10s", V[1]);
      }
      else if (strcmp(V[0], "lon_dms_seconds") == 0)
      {
         sprintf(p_formdata->lon_dms_seconds, "%-10.10s", V[1]);
      }
      else if (strcmp(V[0], "Maidenhead") == 0)
      {
         sprintf(p_formdata->Maidenhead, "%-12.12s", V[1]);
      }
      else if (strcmp(V[0], "Calc_deg") == 0)
      {
         p_formdata->Algor = ALGOR_deg;
      }
      else if (strcmp(V[0], "Calc_dmm") == 0)
      {
         p_formdata->Algor = ALGOR_dmm;
      }
      else if (strcmp(V[0], "Calc_dms") == 0)
      {
         p_formdata->Algor = ALGOR_dms;
      }
      else if (strcmp(V[0], "Calc_Maidenhead") == 0)
      {
         p_formdata->Algor = ALGOR_mdh;
      }

      else if (strcmp(V[0], "NorthSouth") == 0)
      {
         sprintf(p_formdata->NorthSouth, "%-1.1s", V[1]);
      }
      else if (strcmp(V[0], "EastWest") == 0)
      {
         sprintf(p_formdata->EastWest, "%-1.1s", V[1]);
      }
   }

   free(buf);
   if (Vars)
      free(Vars);
}


int initialise (struct formdata_s *p_formdata)
{
   if (!p_formdata->Maidenhead[0])
   {
      memset(p_formdata, '\0', sizeof *p_formdata);

      strcpy(p_formdata->NorthSouth, "N");
      strcpy(p_formdata->EastWest  , "W");

      strcpy(p_formdata->lat_dms_degrees, "41");
      strcpy(p_formdata->lat_dms_minutes, "6");
      strcpy(p_formdata->lat_dms_seconds, "37.0");

      strcpy(p_formdata->lon_dms_degrees, "7");
      strcpy(p_formdata->lon_dms_minutes, "49");
      strcpy(p_formdata->lon_dms_seconds, "59.0");
   }
   return 0;
}

int calculate (struct formdata_s *p_formdata)
{
   double lat_deg_degrees, lat_dmm_minutes, lat_dms_seconds;
   double lon_deg_degrees, lon_dmm_minutes, lon_dms_seconds;

   int lat_dmm_degrees, lat_dms_degrees, lat_dms_minutes;
   int lon_dmm_degrees, lon_dms_degrees, lon_dms_minutes;

   int North = (p_formdata->NorthSouth[0] == 'N');
   int East  = (p_formdata->EastWest  [0] == 'E');

   switch (p_formdata->Algor)
   {
      case ALGOR_deg:
         /* Input is decimal degrees */
         lat_deg_degrees = atof(p_formdata->lat_deg_degrees);
         lon_deg_degrees = atof(p_formdata->lon_deg_degrees);

         lat_deg_degrees *= FlipMinus(lat_deg_degrees, &North);
         lon_deg_degrees *= FlipMinus(lon_deg_degrees, &East );
         lat_deg_degrees = Minimum(lat_deg_degrees,  90);
         lon_deg_degrees = Minimum(lon_deg_degrees, 180);

         /* Calculate degrees and decimal minutes */
         lat_dmm_degrees = (int)(lat_deg_degrees + 0.000005);
         lon_dmm_degrees = (int)(lon_deg_degrees + 0.000005);
         lat_dmm_minutes = (lat_deg_degrees - lat_dmm_degrees) * 60.0;
         lon_dmm_minutes = (lon_deg_degrees - lon_dmm_degrees) * 60.0;

         /* Calculate degrees, minutes and decimal minutes */
         lat_dms_degrees = lat_dmm_degrees;
         lon_dms_degrees = lon_dmm_degrees;
         lat_dms_minutes = (int)(lat_dmm_minutes + 0.000005);
         lon_dms_minutes = (int)(lon_dmm_minutes + 0.000005);
         lat_dms_seconds = fabs(lat_dmm_minutes - lat_dms_minutes) * 60.0;
         lon_dms_seconds = fabs(lon_dmm_minutes - lon_dms_minutes) * 60.0;

         break;

      case ALGOR_dmm:
         /* Input is degrees and decimal minutes */
         lat_dmm_degrees = atof(p_formdata->lat_dmm_degrees);
         lon_dmm_degrees = atof(p_formdata->lon_dmm_degrees);
         lat_dmm_minutes = atof(p_formdata->lat_dmm_minutes);
         lon_dmm_minutes = atof(p_formdata->lon_dmm_minutes);

         lat_dmm_degrees *= FlipMinus(lat_dmm_degrees, &North);
         lon_dmm_degrees *= FlipMinus(lon_dmm_degrees, &East );
         lat_dmm_degrees = Minimum(lat_dmm_degrees,  90);
         lon_dmm_degrees = Minimum(lon_dmm_degrees, 180);

         /* Calculate decimal degrees */
         lat_deg_degrees = lat_dmm_degrees + lat_dmm_minutes / 60.0;
         lon_deg_degrees = lon_dmm_degrees + lon_dmm_minutes / 60.0;

         /* Calculate degrees, minutes and decimal minutes */
         lat_dms_degrees = lat_dmm_degrees;
         lon_dms_degrees = lon_dmm_degrees;
         lat_dms_minutes = (int)(lat_dmm_minutes + 0.000005);
         lon_dms_minutes = (int)(lon_dmm_minutes + 0.000005);
         lat_dms_seconds = fabs(lat_dmm_minutes - lat_dms_minutes) * 60.0;
         lon_dms_seconds = fabs(lon_dmm_minutes - lon_dms_minutes) * 60.0;

         break;

      case ALGOR_dms:
         /* Input is degrees, minutes and decimal minutes */
         lat_dms_degrees = atof(p_formdata->lat_dms_degrees);
         lon_dms_degrees = atof(p_formdata->lon_dms_degrees);
         lat_dms_minutes = atof(p_formdata->lat_dms_minutes);
         lon_dms_minutes = atof(p_formdata->lon_dms_minutes);
         lat_dms_seconds = atof(p_formdata->lat_dms_seconds);
         lon_dms_seconds = atof(p_formdata->lon_dms_seconds);

         lat_dms_degrees *= FlipMinus(lat_dms_degrees, &North);
         lon_dms_degrees *= FlipMinus(lon_dms_degrees, &East );
         lat_dms_degrees = Minimum(lat_dms_degrees,  90);
         lon_dms_degrees = Minimum(lon_dms_degrees, 180);

         /* Calculate decimal degrees */
         lat_deg_degrees = lat_dms_degrees + lat_dms_minutes / 60.0 + lat_dms_seconds / 3600.0;
         lon_deg_degrees = lon_dms_degrees + lon_dms_minutes / 60.0 + lon_dms_seconds / 3600.0;

         /* Calculate degrees and decimal minutes */
         lat_dmm_degrees = lat_dms_degrees;
         lon_dmm_degrees = lon_dms_degrees;
         lat_dmm_minutes = lat_dms_minutes + lat_dms_seconds / 60.0;
         lon_dmm_minutes = lon_dms_minutes + lon_dms_seconds / 60.0;

         break;

      case ALGOR_mdh:
         /* Input is Maidenhead Locator */
         ValidateMaidenhead(p_formdata);

         /* Calculate */
         lon_deg_degrees  = (p_formdata->Maidenhead[0] - 'J') * Maidenhead_factors[0];
         lat_deg_degrees  = (p_formdata->Maidenhead[1] - 'J') * Maidenhead_factors[1];

         lon_deg_degrees += (p_formdata->Maidenhead[2] - '0') * Maidenhead_factors[2];
         lat_deg_degrees += (p_formdata->Maidenhead[3] - '0') * Maidenhead_factors[3];

         lon_deg_degrees += Maidenhead_factors[4] *
            (!p_formdata->Maidenhead[4] ? 12 : p_formdata->Maidenhead[4] - 'a');
         lat_deg_degrees += Maidenhead_factors[5] *
            (!p_formdata->Maidenhead[5] ? 12 : p_formdata->Maidenhead[5] - 'a');

         lon_deg_degrees += Maidenhead_factors[6] *
            (!p_formdata->Maidenhead[6] ?  5 : p_formdata->Maidenhead[6] - '0');
         lat_deg_degrees += Maidenhead_factors[7] *
            (!p_formdata->Maidenhead[7] ?  5 : p_formdata->Maidenhead[7] - '0');

         lon_deg_degrees += Maidenhead_factors[8] *
            (!p_formdata->Maidenhead[8] ? 12 : p_formdata->Maidenhead[8] - 'a');
         lat_deg_degrees += Maidenhead_factors[9] *
            (!p_formdata->Maidenhead[9] ? 12 : p_formdata->Maidenhead[9] - 'a');

         lon_deg_degrees += Maidenhead_factors[10] *
            (!p_formdata->Maidenhead[10] ?  5 : p_formdata->Maidenhead[10] - '0');
         lat_deg_degrees += Maidenhead_factors[11] *
            (!p_formdata->Maidenhead[11] ?  5 : p_formdata->Maidenhead[11] - '0');

         p_formdata->NorthSouth[0] = 'N';
         p_formdata->EastWest  [0] = 'E';
         North = East = 1;

         lat_deg_degrees *= FlipMinus(lat_deg_degrees, &North);
         lon_deg_degrees *= FlipMinus(lon_deg_degrees, &East );
         lat_deg_degrees = Minimum(lat_deg_degrees,  90);
         lon_deg_degrees = Minimum(lon_deg_degrees, 180);

         /* Calculate degrees and decimal minutes */
         lat_dmm_degrees = (int)(lat_deg_degrees + 0.000005);
         lon_dmm_degrees = (int)(lon_deg_degrees + 0.000005);
         lat_dmm_minutes = (lat_deg_degrees - lat_dmm_degrees) * 60.0;
         lon_dmm_minutes = (lon_deg_degrees - lon_dmm_degrees) * 60.0;

         /* Calculate degrees, minutes and decimal minutes */
         lat_dms_degrees = lat_dmm_degrees;
         lon_dms_degrees = lon_dmm_degrees;
         lat_dms_minutes = (int)(lat_dmm_minutes + 0.000005);
         lon_dms_minutes = (int)(lon_dmm_minutes + 0.000005);
         lat_dms_seconds = fabs(lat_dmm_minutes - lat_dms_minutes) * 60.0;
         lon_dms_seconds = fabs(lon_dmm_minutes - lon_dms_minutes) * 60.0;

         break;
   }

   sprintf(p_formdata->lat_deg_degrees, "%13.7f", lat_deg_degrees);

   sprintf(p_formdata->lon_deg_degrees, "%13.7f", lon_deg_degrees);

   sprintf(p_formdata->lat_dmm_degrees, "%3d"  , lat_dmm_degrees);
   sprintf(p_formdata->lat_dmm_minutes, "%9.5f", lat_dmm_minutes);

   sprintf(p_formdata->lon_dmm_degrees, "%3d"  , lon_dmm_degrees);
   sprintf(p_formdata->lon_dmm_minutes, "%9.5f", lon_dmm_minutes);

   sprintf(p_formdata->lat_dms_degrees, "%3d"  , lat_dms_degrees);
   sprintf(p_formdata->lat_dms_minutes, "%3d"  , lat_dms_minutes);
   sprintf(p_formdata->lat_dms_seconds, "%7.3f", lat_dms_seconds);

   sprintf(p_formdata->lon_dms_degrees, "%3d"  , lon_dms_degrees);
   sprintf(p_formdata->lon_dms_minutes, "%3d"  , lon_dms_minutes);
   sprintf(p_formdata->lon_dms_seconds, "%7.3f", lon_dms_seconds);

   sprintf(p_formdata->NorthSouth, North ? "N" : "S");
   sprintf(p_formdata->EastWest  , East  ? "E" : "W");

   if (p_formdata->Algor != ALGOR_mdh)
      CalculateMaidenhead(p_formdata);

   sprintf(p_formdata->GoogleMapsUrl,
      "https://google.com/maps/@%.9f,%.9f,18z",
      lat_deg_degrees * (North ? 1.0 : -1.0),
      lon_deg_degrees * (East  ? 1.0 : -1.0));

   return 0;
}

int FlipMinus (double degrees, int *p_dir)
{
   if (degrees < 0)
   {
      *p_dir = !*p_dir;
      return -1;
   }
   return 1;
}

void ValidateMaidenhead (struct formdata_s *p_formdata)
{
   /* Formal checks and corrections */
   if (!isalpha(p_formdata->Maidenhead[0]))
      p_formdata->Maidenhead[0] = 'J';
   else
      p_formdata->Maidenhead[0] = toupper(p_formdata->Maidenhead[0]);

   if (!isalpha(p_formdata->Maidenhead[1]))
      p_formdata->Maidenhead[1] = 'J';
   else
      p_formdata->Maidenhead[1] = toupper(p_formdata->Maidenhead[1]);

   if (p_formdata->Maidenhead[0] > 'R')
       p_formdata->Maidenhead[0] = 'R';
   if (p_formdata->Maidenhead[1] > 'R')
       p_formdata->Maidenhead[1] = 'R';

   if (!isdigit(p_formdata->Maidenhead[2]))
       p_formdata->Maidenhead[2] = '0';
   if (!isdigit(p_formdata->Maidenhead[3]))
       p_formdata->Maidenhead[3] = '0';

   if (!isalpha(p_formdata->Maidenhead[4]) ||
       !isalpha(p_formdata->Maidenhead[5]))
      memset(p_formdata->Maidenhead + 4, '\0', sizeof p_formdata->Maidenhead - 4);
   else
   {
      p_formdata->Maidenhead[4] = tolower(p_formdata->Maidenhead[4]);
      p_formdata->Maidenhead[5] = tolower(p_formdata->Maidenhead[5]);
   }

   if (p_formdata->Maidenhead[4] > 'x')
       p_formdata->Maidenhead[4] = 'x';
   if (p_formdata->Maidenhead[5] > 'x')
       p_formdata->Maidenhead[5] = 'x';

   if (!isdigit(p_formdata->Maidenhead[6]) ||
       !isdigit(p_formdata->Maidenhead[7]))
   {
       p_formdata->Maidenhead[6] = '\0';
       p_formdata->Maidenhead[7] = '\0';
   }

   if (!isalpha(p_formdata->Maidenhead[8]) ||
       !isalpha(p_formdata->Maidenhead[9]))
      memset(p_formdata->Maidenhead + 8, '\0', sizeof p_formdata->Maidenhead - 8);
   else
   {
      p_formdata->Maidenhead[8] = tolower(p_formdata->Maidenhead[8]);
      p_formdata->Maidenhead[9] = tolower(p_formdata->Maidenhead[9]);
   }

   if (p_formdata->Maidenhead[8] > 'x')
       p_formdata->Maidenhead[8] = 'x';
   if (p_formdata->Maidenhead[9] > 'x')
       p_formdata->Maidenhead[9] = 'x';

   if (!isdigit(p_formdata->Maidenhead[10]) ||
       !isdigit(p_formdata->Maidenhead[11]))
   {
       p_formdata->Maidenhead[10] = '\0';
       p_formdata->Maidenhead[11] = '\0';
   }
}

void CalculateMaidenhead (struct formdata_s *p_formdata)
{
   double lon_deg = atof(p_formdata->lon_dms_degrees) +
                    atof(p_formdata->lon_dms_minutes) / 60.0 +
                    atof(p_formdata->lon_dms_seconds) / 3600.0;
   double lat_deg = atof(p_formdata->lat_dms_degrees) +
                    atof(p_formdata->lat_dms_minutes) / 60.0 +
                    atof(p_formdata->lat_dms_seconds) / 3600.0;

   memset(p_formdata->Maidenhead, '\0', sizeof p_formdata->Maidenhead);

   if (p_formdata->EastWest  [0] == 'E')
      lon_deg += 180.0;
   else
      lon_deg  = 180.0 - lon_deg;
   if (p_formdata->NorthSouth[0] == 'N')
      lat_deg +=  90.0;
   else
      lat_deg  =  90.0 - lat_deg;

   p_formdata->Maidenhead[0] = (int)(lon_deg / Maidenhead_factors[0]) + 'A';
   lon_deg -= (p_formdata->Maidenhead[0] - 'A') * Maidenhead_factors[0];

   p_formdata->Maidenhead[1] = (int)(lat_deg / Maidenhead_factors[1]) + 'A';
   lat_deg -= (p_formdata->Maidenhead[1] - 'A') * Maidenhead_factors[1];

   p_formdata->Maidenhead[2] = (int)(lon_deg / Maidenhead_factors[2]) + '0';
   lon_deg -= (p_formdata->Maidenhead[2] - '0') * Maidenhead_factors[2];

   p_formdata->Maidenhead[3] = (int)(lat_deg / Maidenhead_factors[3]) + '0';
   lat_deg -= (p_formdata->Maidenhead[3] - '0') * Maidenhead_factors[3];

   p_formdata->Maidenhead[4] = (int)(lon_deg / Maidenhead_factors[4]) + 'a';
   lon_deg -= (p_formdata->Maidenhead[4] - 'a') * Maidenhead_factors[4];

   p_formdata->Maidenhead[5] = (int)(lat_deg / Maidenhead_factors[5]) + 'a';
   lat_deg -= (p_formdata->Maidenhead[5] - 'a') * Maidenhead_factors[5];

   p_formdata->Maidenhead[6] = (int)(lon_deg / Maidenhead_factors[6]) + '0';
   lon_deg -= (p_formdata->Maidenhead[6] - '0') * Maidenhead_factors[6];

   p_formdata->Maidenhead[7] = (int)(lat_deg / Maidenhead_factors[7]) + '0';
   lat_deg -= (p_formdata->Maidenhead[7] - '0') * Maidenhead_factors[7];

   p_formdata->Maidenhead[8] = (int)(lon_deg / Maidenhead_factors[8]) + 'a';
   lon_deg -= (p_formdata->Maidenhead[8] - 'a') * Maidenhead_factors[8];

   p_formdata->Maidenhead[9] = (int)(lat_deg / Maidenhead_factors[9]) + 'a';
   lat_deg -= (p_formdata->Maidenhead[9] - 'a') * Maidenhead_factors[9];

   p_formdata->Maidenhead[10] = (int)(lon_deg / Maidenhead_factors[10] + 0.5) + '0';
   if (p_formdata->Maidenhead[10] > '9') p_formdata->Maidenhead[10] = '9';
   lon_deg -= (p_formdata->Maidenhead[10] - '0') * Maidenhead_factors[10];

   p_formdata->Maidenhead[11] = (int)(lat_deg / Maidenhead_factors[11] + 0.5) + '0';
   if (p_formdata->Maidenhead[11] > '9') p_formdata->Maidenhead[11] = '9';
   lat_deg -= (p_formdata->Maidenhead[11] - '0') * Maidenhead_factors[11];

   /* Special case North pole, exactly 90 degrees North. */
   if (p_formdata->Maidenhead[1] > 'R')
   {
      p_formdata->Maidenhead[1] = 'R';
      p_formdata->Maidenhead[3] = '9';
      p_formdata->Maidenhead[5] = 'x';
      p_formdata->Maidenhead[7] = '9';
      p_formdata->Maidenhead[9] = 'x';
      p_formdata->Maidenhead[11]= '9';
   }
   /* Special case exactly 180 degrees East or West) */
   if (p_formdata->Maidenhead[0] > 'R')
   {
      p_formdata->Maidenhead[0] = 'R';
      p_formdata->Maidenhead[2] = '9';
      p_formdata->Maidenhead[4] = 'x';
      p_formdata->Maidenhead[6] = '9';
      p_formdata->Maidenhead[8] = 'x';
      p_formdata->Maidenhead[10]= '9';
   }

   /* Remove superfluous codes from the end, which only indicate the
      centre of the square. */
   if (strcmp(p_formdata->Maidenhead + 10, "55") == 0)
       memset(p_formdata->Maidenhead + 10, '\0', sizeof p_formdata->Maidenhead - 10);
   if (strcmp(p_formdata->Maidenhead +  8, "mm") == 0)
       memset(p_formdata->Maidenhead +  8, '\0', sizeof p_formdata->Maidenhead -  8);
   if (strcmp(p_formdata->Maidenhead +  6, "55") == 0)
       memset(p_formdata->Maidenhead +  6, '\0', sizeof p_formdata->Maidenhead -  6);
}
